#!/usr/bin/env python
import numpy as np
import time
import schedule
import actionlib
import rospy
import math
import os
from datetime import datetime
from move_base_msgs.msg import MoveBaseAction, MoveBaseGoal
from actionlib_msgs.msg import *
from geometry_msgs.msg import Pose, PoseWithCovarianceStamped, Point, Quaternion, Twist
from kobuki_msgs.msg import SensorState, AutoDockingAction, AutoDockingGoal, ButtonEvent

class SchedulePosition():
	#VARIABLES
        home_x = 0.671976923943 #coordenada x proxima a la estacion
        home_y = 0.73062390089 #coordenada y proxima a la estacion
        max_carga = 162 
	battery_dangerous = 133 #A partir de este valor la bateria es critica (5%)
        #Inicializar
        LowBattery = False
        DockStation = True
        nivel_bateria_anterior = 1000 #Bateria inalcanzable para iniciar
	home = False
	intentos = 0
	ficheromodificado = False

	def __init__(self):

                self.goal_sent = False

                rospy.on_shutdown(self.shutdown)

                self.move_base = actionlib.SimpleActionClient("move_base", MoveBaseAction)
                rospy.loginfo("Esperando al servidor move_base action para empezar")

                if(self.move_base.wait_for_server()):
			rospy.loginfo("Servidor encontrado")

		rospy.Subscriber("/mobile_base/sensors/core",SensorState,self.SensorStateCallback)
		rospy.Subscriber("/mobile_base/events/button",ButtonEvent,self.ButtonEventCallback)
		rospy.sleep(1)

	def shutdown(self):
                if self.goal_sent:
                        self.move_base.cancel_goal()
                rospy.loginfo("Stop")
                rospy.sleep(1)

	def LeerFichero(self):
        	Schedule = []
        	tarea = []
        	f = open("/home/tb2/catkin_ws/src/scheduleposition/scripts/Schedule.txt")
        	lineas = f.readlines() #lineas es una list con cada linea del fichero en cada posicion
		f.close()
        	for line in lineas:
            		frase = line.strip() #Borra espacios y \n al principio y al final del string
			if (frase != ""): #Si no es una linea en blanco, guarda la tarea. Para evitar errores si hay lineas en blanco en el fichero
            			frase = frase.replace(" ","") #Por si se ponen espacio entre medias, para evitar errores, los borra (si hay dos espacios seguidos no los borra!!)
            			hora = frase.partition(";")[0]
            			posicion = frase.partition(";")[2]
				if (posicion == "home"):
					x = self.home_x
					y = self.home_y
				else:
            				x = float(posicion.partition(";")[0])
            				y = float(posicion.partition(";")[2])
              			tarea.append(hora)
 				tarea.append(x)
 	   			tarea.append(y)   
           			Schedule.append(tarea) #Matriz de tareas
           			tarea = [] #Una vez guardada la tarea, se borra para meter una nueva
        	return Schedule

	def ult_modificacion(self):
		estado = os.stat("/home/tb2/catkin_ws/src/scheduleposition/scripts/Schedule.txt")
		fecha = time.localtime(estado.st_mtime)
		fecha = datetime(fecha[0], fecha[1], fecha[2], fecha[3], fecha[4], fecha[5])
		return fecha
	
	def escribe(pos, quat):
		print(pos)
		print(quat)
		time.sleep(5)

	def goto(self, pos, quat):
		if pos['x'] == self.home_x and pos['y'] == self.home_y and self.DockStation:
			rospy.loginfo("Ya estamos en la base")
			result = True
		else:
			self.BackUp_DockStation()
			rospy.loginfo("Go to (%s, %s) pose", pos['x'], pos['y'])
			self.goal_sent = True
        		goal = MoveBaseGoal()
       			goal.target_pose.header.frame_id = 'map'
        		goal.target_pose.header.stamp = rospy.Time.now()
        		goal.target_pose.pose = Pose(Point(pos['x'], pos['y'], 0.000), Quaternion(quat['r1'], quat['r2'], quat['r3'], quat['r4']))

        		self.move_base.send_goal(goal)

        		success = self.move_base.wait_for_result(rospy.Duration(60))

        		state = self.move_base.get_state()
        		result = False
     			if success and state == GoalStatus.SUCCEEDED:
				result = True
				if pos['x'] == self.home_x and pos['y'] == self.home_y:
					self.home = True
					self.GoToDockStation()
        		else:
                		self.move_base.cancel_goal()
				cmd_vel = rospy.Publisher('cmd_vel_mux/input/navi', Twist, queue_size=10)
				cmd_vel.publish(Twist())
				rospy.loginfo("Imposible llegar al punto deseado")
				if (self.intentos < 3):
					self.GoToDockStation()
					self.intentos+=1
					print(self.intentos)
				else:
					sys.exit("error")
					self.move_base.cancel_goal()
					cmd_vel.publish(Twist())
						
		return result

	def main(self, horario):
		i = 0
		while i<len(horario):
			position = {'x': horario[i][1], 'y': horario[i][2]}
			quaternion = {'r1' : 0.000, 'r2' : 0.000, 'r3' : 0.000, 'r4' : 1.000}
			schedule.every().day.at(horario[i][0]).do(self.goto, position, quaternion)
			i = i+1

	def SensorStateCallback(self, data):
                #data.battery esta variando continuamente entre dos valores, luego solo avisa  si hay una diferencia de mas de 2
                if(math.fabs(int(data.battery) - self.nivel_bateria_anterior) > 2):
			rospy.loginfo("Nivel de bateria: " + str(round((float(data.battery)-float(self.battery_dangerous)) / (float(self.max_carga)-float(self.battery_dangerous)) * 100)) + "%")
			self.nivel_bateria_anterior = int(data.battery)

                if(int(data.charger) == 0): #Si es un 0, significa que se esta descargando, luego no esta cargando.
                        self.DockStation = False
                else:
                        self.DockStation = True

		#Si la bateria disminuye por debajo del 25% de carga, significa que la bateria es baja
                if (round((float(data.battery)-float(self.battery_dangerous)) / (float(self.max_carga)-float(self.battery_dangerous)) * 100) < 40) : 
                        if(not self.LowBattery):
                                rospy.loginfo("BATERIA BAJA")
                        self.LowBattery = True
		#Hasta que no este por encima del 35% no dejamos que salga de la estacion
		#Asi evitamos que se ponga a cargar y al instante salga
                if ((round(float(data.battery)-float(self.battery_dangerous)) / (float(self.max_carga)-float(self.battery_dangerous)) * 100) > 50) :
                        if(self.LowBattery):
                                rospy.loginfo("Puede salir de la estacion")     
                        self.LowBattery = False

	def BackUp_DockStation(self):
                if(self.DockStation == True):
                        rospy.loginfo("Robot en estacion. Dando marcha atras antes de empezar")
                        self.DockStation = False
			self.home = False
                        cmd_vel = rospy.Publisher('cmd_vel_mux/input/navi', Twist, queue_size=10)
                        #Twist es el tipo de mensaje
                        move_cmd = Twist()
                        #Marcha atras a  0.1 m/s
                        move_cmd.linear.x = -0.1
                        move_cmd.angular.z = 0

                        r = rospy.Rate(10);
                        temp_count = 0
                        #Marcha atras durante 30 segundos
                        while (not rospy.is_shutdown() and temp_count < 300):
                                cmd_vel.publish(move_cmd)
                                #Espera hasta 0.1 s (10 HZ) y publica otra vez
                                temp_count = temp_count + 1
                                r.sleep()
                        #Nos aseguramos de que para publicando un Twist()
                        cmd_vel.publish(Twist())
                        return True

	def GoToDockStation(self):
		#Primero nos aproximamos a la base
		#home = False
		pos = {'x': self.home_x, 'y': self.home_y}
                quat = {'r1' : 0.000, 'r2' : 0.000, 'r3' : 0.000, 'r4' : 1.000}
		self.home = self.goto(pos, quat)

		#Tras esto, aparcamos
		self._client = actionlib.SimpleActionClient('/dock_drive_action', AutoDockingAction)
                rospy.loginfo("Esperando al servidor auto_docking")
                self._client.wait_for_server()
                rospy.loginfo("Servidor auto_docking encontrado")

                goal = AutoDockingGoal()
                rospy.loginfo("Sending auto_docking goal and waiting for result (times out in 180 seconds and will try again if required)")
                self._client.send_goal(goal)

                #Le damos 180seg para aparcar
                success = self._client.wait_for_result(rospy.Duration(180))

                if success:
                        rospy.loginfo("Aparcamiento conseguido")
                        self.DockStation = True #Estamos en la base
			self.intentos = 0
                        return True
                else:
                        rospy.loginfo("Fallo en el aparcamiento")
                        return False

	def NeedCharge(self):
		#Si la bateria es baja pero estamos en la base, nos quedamos  hasta que este suficientemente cargado
                if(self.DockStation and self.LowBattery):
                        rospy.loginfo("Robot en la base de carga")
                        rospy.loginfo("Esperando a que este suficientemente cargado")
                        time.sleep(30)
                        return True

                #Si la bateria es baja y no estamos en la base, va a ella
                if(not self.DockStation and self.LowBattery):
                        rospy.loginfo("Bateria baja. Yendo a la base de carga")
                        self.GoToDockStation()
                        return True
                return False

	def ButtonEventCallback(self, data):
		if (data.state == ButtonEvent.PRESSED):
			if (data.button == ButtonEvent.Button0):
				if (schedule.idle_seconds() <= 180): #Segundos hasta la siguiente tarea
					#Esto lo hago porque las tareas del schedule entran al programa parando cualquier cosa
					#que este haciendo.  Si esta yendo a la base y salta una tarea, parara y realizara la tarea
					rospy.loginfo("QUEDAN MENOS DE 3 MINUTOS PARA LA SIGUIENTE TAREA")
					rospy.loginfo("HASTA QUE NO LA REALICE, NO PUEDE IR A HOME POR SEGURIDAD")
				else:
					if(not self.DockStation):
						rospy.loginfo("Yendo a la base...")
						self.GoToDockStation()
					else:
						rospy.loginfo("Ya estamos en la base")

if __name__ == '__main__':
        try:
		rospy.init_node('SchedulePosition', anonymous=False)
               	navigator = SchedulePosition()
                horario = navigator.LeerFichero()
		primeramod =  navigator.ult_modificacion()
		navigator.main(horario)
                while not rospy.is_shutdown():
			ultmod = navigator.ult_modificacion()
			if (ultmod != primeramod):
				navigator.ficheromodificado = True
				primeramod = ultmod
				print("Fichero modificado")
			else:
				navigator.ficheromodificado = False
			if (not navigator.NeedCharge() and not navigator.ficheromodificado):
                       		schedule.run_pending()
			else:
				print("Actualizando horario")
				schedule.clear()
				horario = navigator.LeerFichero()
				navigator.main(horario)
			time.sleep(1)


        except rospy.ROSInterruptException:
                rospy.loginfo("Ctrl-C caught. Quitting")
