#!/usr/bin/env python
import numpy as np
import time
import schedule
import actionlib
import rospy
import math
import os
from datetime import datetime
from std_srvs.srv import *
from move_base_msgs.msg import MoveBaseAction, MoveBaseGoal
from actionlib_msgs.msg import *
from geometry_msgs.msg import Pose, PoseWithCovarianceStamped, Point, Quaternion, Twist
from kobuki_msgs.msg import SensorState, AutoDockingAction, AutoDockingGoal, ButtonEvent, Sound

class SchedulePosition():
	#VARIABLES
    home_x = 0.671976923943 #coordenada x proxima a la estacion
    home_y = 0.73062390089 #coordenada y proxima a la estacion
    max_carga = 162 #Bateria al 100%
	battery_dangerous = 133 #A partir de este valor la bateria es critica (5%)
    #Inicializar
    LowBattery = False #Parametro indicador de poca bateria
    DockStation = True #Parametro indicador de que el robot se encuentra en la estacion
    nivel_bateria_anterior = 1000 #Bateria inalcanzable para iniciar
	home = False #Parametro indicador de que el  robot se encuentra cerca de la estacion para aparcar 
	intentos = 0
	ficheromodificado = False

#Funcion constructor
	def __init__(self):
        self.goal_sent = False #Borra cualquier destino enviado al iniciar
        rospy.on_shutdown(self.shutdown) #Funcion a la que se llama al recibir CTL+C
        self.move_base = actionlib.SimpleActionClient("move_base", MoveBaseAction) #Se crea el cliente al servidor MoveBaseAction
        rospy.loginfo("Esperando al servidor move_base action para empezar")
        if(self.move_base.wait_for_server()): #Espera a encontrar el servidor para continuar, por si no se ha ejecutado antes 
            rospy.loginfo("Servidor encontrado")
		self.sound = rospy.Publisher('/mobile_base/commands/sound', Sound, queue_size=10) #Publicador encargado de sonidos
		rospy.Subscriber("/mobile_base/sensors/core",SensorState,self.SensorStateCallback) #Subscriptor para sensores de bateria 
		rospy.Subscriber("/mobile_base/events/button",ButtonEvent,self.ButtonEventCallback) #Subscriptor para los botones del robot 
		rospy.sleep(1) #Para que de tiempo a arrancar publicadores y suscriptores

#Funcion llamada al recibir CTL+C
	def shutdown(self):
        if self.goal_sent:
            self.move_base.cancel_goal() #Si se ha enviado un destino, lo cancela
        rospy.loginfo("Ctrl-C caught. Quitting")
		self.ReproducirSonido(1) #Reproduce sonido 
        rospy.sleep(1) #Espera 1 segundo para que le de tiempo a parar

#Funcion para leer del fichero
	def LeerFichero(self):
        Schedule = []
        tarea = []
        f = open("/home/tb2/catkin_ws/src/scheduleposition/scripts/Schedule.txt") #Ruta donde va a estar el fichero
        lineas = f.readlines() #lineas es una list con cada linea del fichero en cada posicion
		f.close()
        for line in lineas:
           	frase = line.strip() #Borra espacios y \n al principio y al final del string
            if (frase != ""): #Si no es una linea en blanco, guarda la tarea. Para evitar errores si hay lineas en blanco en el fichero
           		frase = frase.replace(" ","") #Por si se ponen espacio entre medias, para evitar errores, los borra (si hay dos espacios seguidos no los borra!!)
           		hora = frase.partition(";")[0] #Divide la frase en tres posiciones siendo la 0 la hora, la 1 el ';' y la 2 la posicion  
           		posicion = frase.partition(";")[2]
                if (posicion == "home"): #Si lee home, se vuelve a la estacion
                    x = self.home_x
                    y = self.home_y
                    q0 = float("1")
                    q1 = float("0")
                    q2 = float("0")
                    q3 = float("0")
                else:
          			x = float(posicion.partition(";")[0])
           			y = float(posicion.partition(";")[2].partition(";")[0])
                    orientacion = posicion.partition(";")[2].partition(";")[2]
                    q0 = float(orientacion.partition(";")[0])
                    q1 = float(orientacion.partition(";")[2].partition(";")[0])
                    q2 = float(orientacion.partition(";")[2].partition(";")[2].partition(";")[0])
                    q3 = float(orientacion.partition(";")[2].partition(";")[2].partition(";")[2]) #Se va partiendo el string para dividir cada parte en una variable
           		tarea.append(hora)
                tarea.append(x)
                tarea.append(y)
                tarea.append(q0)
                tarea.append(q1)
                tarea.append(q2)
                tarea.append(q3) 
       			Schedule.append(tarea) #Matriz de tareas
       			tarea = [] #Una vez guardada la tarea, se borra para meter una nueva
       	return Schedule

	def ult_modificacion(self):
		estado = os.stat("/home/tb2/catkin_ws/src/scheduleposition/scripts/Schedule.txt")
		fecha = time.localtime(estado.st_mtime)
		fecha = datetime(fecha[0], fecha[1], fecha[2], fecha[3], fecha[4], fecha[5])
		return fecha

	def goto(self, pos, quat):
		if not pos['x'] == self.home_x and not pos['y'] == self.home_y: #Si el destino no es la base, si esta en la base da marcha atras
			self.BackUp_DockStation()
		self.BorrarCostmaps() #Para borrar obstaculos que pueden haber desaparecido
		rospy.loginfo("Go to (%s, %s) pose", pos['x'], pos['y'])
		rospy.loginfo("Quaternion (%s, %s, %s, %s) ",quat['r1'], quat['r2'], quat['r3'], quat['r4'])
		self.goal_sent = True
        goal = MoveBaseGoal()
       	goal.target_pose.header.frame_id = 'map'
        goal.target_pose.header.stamp = rospy.Time.now()
        goal.target_pose.pose = Pose(Point(pos['x'], pos['y'], 0.000), Quaternion(quat['r1'], quat['r2'], quat['r3'], quat['r4'])) #destino al que hay que ir
        if pos['x'] == self.home_x and pos['y'] == self.home_y: 
            result = self.GoToDockStation() #si el destino es la base, se llama a la funcion encargada de llevar a la base 
        else: #si el destino no es la base 
			self.home = False #sale de la base asi que se cambia el valor de esta variable 
        	self.move_base.send_goal(goal) #Manda el destino 
	       	success = self.move_base.wait_for_result(rospy.Duration(500)) #Da 500 segundos para llegar 
            state = self.move_base.get_state() #Devuelve el estado de la trayectoria (si esta atascado, lo ha conseguido...)
            result = False
            if success and state == GoalStatus.SUCCEEDED: #Si ha llegado al destino y esta dentro del tiempo devuelve true
                result = True
            else:
                self.move_base.cancel_goal() #si no llega, cancela el destino
                cmd_vel = rospy.Publisher('cmd_vel_mux/input/navi', Twist, queue_size=10) 
                cmd_vel.publish(Twist()) #publica velocidad 0 para parar
                rospy.loginfo("Imposible llegar al punto deseado") 
                self.ReproducirSonido(4) #reproduce sonido de error 
                rospy.sleep(5) #espera 5 segundos 
                self.GoToDockStation() #vuelve a la estacion 
		return result

#Funcion encargada de meter el horario en el planificador
	def planificar(self, horario):
		i = 0
		while i<len(horario): #mientras haya posiciones en la lista (lineas en el fichero)
			position = {'x': horario[i][1], 'y': horario[i][2]}
			quaternion = {'r1' : horario[i][4], 'r2' : horario[i][5], 'r3' : horario[i][6], 'r4' : horario[i][3]}
			schedule.every().day.at(horario[i][0]).do(self.goto, position, quaternion) #llamamos a la horas indicadas a la funcion goto() que manda a las posiciones indicadas 
			i = i+1

#Funcion a la que se llama cuando se recibe un mensaje de la bateria 
    def SensorStateCallback(self, data):
        #data.battery esta variando continuamente entre dos valores, luego solo actualiza el valor si hay una diferencia de mas de 2
        if(math.fabs(int(data.battery) - self.nivel_bateria_anterior) > 2):
			rospy.loginfo("Nivel de bateria: " + str(round((float(data.battery)-float(self.battery_dangerous)) / (float(self.max_carga)-float(self.battery_dangerous)) * 100)) + "%")
			self.nivel_bateria_anterior = int(data.battery)

        if(int(data.charger) == 0): #Si es un 0, significa que se esta descargando, luego no esta cargando.
            self.DockStation = False #lo que significa que no estaria en la estacion 
        else:
            self.DockStation = True
        
        #Si la bateria disminuye por debajo del 10% de carga, significa que la bateria es baja
        if (round((float(data.battery)-float(self.battery_dangerous)) / (float(self.max_carga)-float(self.battery_dangerous)) * 100) < 10) : 
            if(not self.LowBattery):
                rospy.loginfo("BATERIA BAJA")
                self.LowBattery = True #Bateria baja
		
        #Hasta que no este por encima del 20% no dejamos que salga de la estacion
		#Asi evitamos que se ponga a cargar y al instante salga
        if ((round(float(data.battery)-float(self.battery_dangerous)) / (float(self.max_carga)-float(self.battery_dangerous)) * 100) > 20) :
            if(self.LowBattery):
                rospy.loginfo("Puede salir de la estacion")     
                self.LowBattery = False #Ya no hay bateria baja

#Funcion encargada de dar marcha atras si estamos en la base y recibimos un destino 
    def BackUp_DockStation(self):
        if(self.DockStation == True):
            rospy.loginfo("Robot en estacion. Dando marcha atras antes de empezar")
            self.DockStation = False
            self.home = False
            cmd_vel = rospy.Publisher('cmd_vel_mux/input/navi', Twist, queue_size=10)
            #Twist es el tipo de mensaje
            move_cmd = Twist()
            #Marcha atras a  0.1 m/s
            move_cmd.linear.x = -0.1
            move_cmd.angular.z = 0
            r = rospy.Rate(10); #10Hz
            temp_count = 0
            #Marcha atras durante 30 segundos
            while (not rospy.is_shutdown() and temp_count < 300):
                    cmd_vel.publish(move_cmd)
                    temp_count = temp_count + 1
                    r.sleep() #Espera hasta 0.1 s (10 HZ) y publica otra vez
            #Nos aseguramos de que para publicando un Twist()
            cmd_vel.publish(Twist())
        return True

#Funcion para ir a la base de carga 
    def GoToDockStation(self):
		#Primero nos aproximamos a la base
        if self.DockStation: #Si ya esta en la base no hace nada
            rospy.loginfo("Ya estamos en la base")
            result = True
        else:
            while (self.home == False): #mientras no estemos cerca de la base, intentamos llegar a ese punto cercano 
                if(self.intentos<3): #se intenta 3 veces
                    rospy.loginfo("Going HOME")
                    self.BorrarCostmaps() #se borran mapas de costo
                    goal = MoveBaseGoal()
                    goal.target_pose.header.frame_id = 'map'
                    goal.target_pose.header.stamp = rospy.Time.now()
                    goal.target_pose.pose = Pose(Point(self.home_x, self.home_y, 0.000), Quaternion(0.000,0.000, 0.000, 1.000))
                    self.move_base.send_goal(goal) #se le manda el punto cercano a la base como destino 

                    success = self.move_base.wait_for_result(rospy.Duration(500)) #se le deja 500 segundos para llegar
                    state = self.move_base.get_state()
                    result = False
                    if success and state == GoalStatus.SUCCEEDED: #si llega al punto 
		                result = True
                        self.home = True #indica que estamos cerca de la base 
                        self.intentos = 0 #reinicia intentos 
                    else: #si no llega al punto 
                        rospy.loginfo("ERROR AL LLEGAR A LA BASE, VOLVIENDO A INTENTARLO")
                        self.move_base.cancel_goal() #cancela el destino 
                        cmd_vel = rospy.Publisher('cmd_vel_mux/input/navi', Twist, queue_size=10)
                        cmd_vel.publish(Twist()) #paramos el robot 
                        self.intentos +=1 #incrementamos el valor del intento 
                        self.ReproducirSonido(4) #reproduce sonido de error 
                        rospy.loginfo("Numero intentos: %d", self.intentos) #muestra el numero de intentos que lleva
                else: #si no lo consigue en tres intentos 
           			self.move_base.cancel_goal() #cancela el destino 
                    cmd_vel.publish(Twist()) #para el robot
                    result = False
                    rospy.signal_shutdown("ERROR. Imposible llegar a HOME") #sale del programa
                    
			#Si ha llegado cerca de la base, aparcamos
            self._client = actionlib.SimpleActionClient('/dock_drive_action', AutoDockingAction) #cliente del servidor AutoDockingAction
            rospy.loginfo("Esperando al servidor auto_docking")
            self._client.wait_for_server() #espera hasta encontrar el servidor por si no se ha ejecutado 
            rospy.loginfo("Servidor auto_docking encontrado")
            goal = AutoDockingGoal()
            rospy.loginfo("APARCANDO... (cancelando si pasa de 180s)")
            self._client.send_goal(goal) #envia el destino que es la base de carga 
            
            #Le damos 180seg para aparcar
            success = self._client.wait_for_result(rospy.Duration(180))
            if success:
              	rospy.loginfo("Aparcamiento conseguido")
             	self.DockStation = True #Estamos en la base
              	return True
            else:
				self._client.cancel_goal() #cancela el aparcamiento
                rospy.loginfo("Fallo en el aparcamiento")
                return False

#Funcion que va a comprobar si se necesita ir a la base de carga 
    def NeedCharge(self):
		#Si la bateria es baja pero estamos en la base, nos quedamos  hasta que este suficientemente cargado
        if(self.DockStation and self.LowBattery):
            rospy.loginfo("Robot en la base de carga")
            rospy.loginfo("Esperando a que este suficientemente cargado")
            time.sleep(30)
            return True

        #Si la bateria es baja y no estamos en la base, va a ella
        if(not self.DockStation and self.LowBattery):
            rospy.loginfo("Bateria baja. Yendo a la base de carga")
            self.GoToDockStation()
            return True 
        return False

#Funcion llamada tras mensaje de los botones
    def ButtonEventCallback(self, data):
		if (data.state == ButtonEvent.PRESSED): #Si un boton ha sido presionado
			if (data.button == ButtonEvent.Button0): #Y ha sido el boton 0
				if (schedule.idle_seconds() <= 180): #Segundos hasta la siguiente tarea
					#Esto lo hago porque las tareas del schedule entran al programa parando cualquier cosa
					#que este haciendo.  Si esta yendo a la base y salta una tarea, parara y realizara la tarea
					rospy.loginfo("QUEDAN MENOS DE 3 MINUTOS PARA LA SIGUIENTE TAREA")
					rospy.loginfo("HASTA QUE NO LA REALICE, NO PUEDE IR A HOME POR SEGURIDAD")
				else: #si no hay tareas cerca
					if(not self.DockStation):
						rospy.loginfo("Yendo a la base...")
						self.GoToDockStation() #si no esta en la base, vuelve a ella
					else:
						rospy.loginfo("Ya estamos en la base")

#Funcion encargada de borrar los mapas de costo
    def BorrarCostmaps(self):
		rospy.wait_for_service('/move_base/clear_costmaps')
		clear_costmaps = rospy.ServiceProxy('/move_base/clear_costmaps', Empty)
		if(clear_costmaps.call()): #Borra los mapas
			rospy.loginfo('GLOBAL COSTMAPS BORRADOS')
		else:
			rospy.loginfo('Error al llamar al servicio /move_base/clear_costmaps')
            
#Funcion encargada de reproducir sonidos
    def ReproducirSonido(self, num):
		self.sound.publish(num) #publica el sonido que se le pase como argumento
		rospy.sleep(5)

if __name__ == '__main__':
        try:
            rospy.init_node('SchedulePosition', anonymous=False) #inicia el nodo
            navigator = SchedulePosition() #crea objeto
            horario = navigator.LeerFichero() #guarda en horario lo que hay dentro del fichero en una matriz de matrices
            primeramod =  navigator.ult_modificacion() #guarda el valor de la ultima modificacion del fichero 
            navigator.planificar(horario) #planifica
            while not rospy.is_shutdown(): #mientras no se cierre, entra en bucle
                ultmod = navigator.ult_modificacion()  #comprueba la fecha actual de la ultima modificacion del fichero 
                if (ultmod != primeramod): #si es diferente con la guardada anteriormente
                    navigator.ficheromodificado = True #el fichero ha sido modficiado
                    primeramod = ultmod #guardamos la ultima fecha de modificacion 
                    print("Fichero modificado")
                else:
                    navigator.ficheromodificado = False #sino, no ha sido modificado 
                if (not navigator.NeedCharge() and not navigator.ficheromodificado): #si no necesita carga y el fichero no ha sido modificado.
                    schedule.run_pending() #comienza a realizar tareas en funcion de la planificacion
                else:  #si el fichero ha sido modificado o necesita bateria
                    print("Actualizando horario")
                    schedule.clear() #borra el plan anterior 
                    horario = navigator.LeerFichero() #lee el fichero 
                    navigator.planificar(horario) #vuelve a planificarlo
                time.sleep(1)


        except rospy.ROSInterruptException:
                rospy.loginfo("Ctrl-C caught. Quitting")
